package Irpg::Quest;

use strict;
use warnings;
use Irpg::Utils qw(:text);
use Irpg::Irc qw(:interaction);
use Exporter qw(import);
our @EXPORT = qw(&writequestfile &quest &questpencheck &movequesters);


my %quest = (
    questers => [],
    p1 => [], # point 1 for q2
    p2 => [], # point 2 for q2
    qtime => time() + int(rand(21600)), # first quest starts in <=6 hours
    text => "",
    type => 1,
    stage => 1); # quest info

my $opts;
my $rps;
=head1 FUNCTION init_hashes
    This function sets the references to
    options and players hashes, and init quest if 
=over
=item SCALAR (ref)    - reference to the options hash
=item SCALAR (ref)    - reference to the players hash
=cut
sub init_pkg { ($opts, $rps) = @_; readquestfile();}

=head FUNCTION readquestfile
    This function reads a quest state from the file
    and set it.
=cut
sub readquestfile {
    return unless $opts->{writequestfile};
    open(QF,"<$opts->{questfilename}") or do {
        Irpg::Irc::chanmsg("Error: Cannot open $opts->{questfilename}: $!");
        return;
    };
    while ( <QF> ) {
        chomp;
        $quest{text}  = substr($_, 2) if (m/^T /);
        $quest{type}  = substr($_, 2) if (m/^Y /);
        $quest{qtime} = substr($_, 2) if (m/^S / && $quest{type} == 1);
        $quest{stage} = substr($_, 2) if (m/^S / && $quest{type} == 2);
        (
            $quest{p1}->[0],
            $quest{p1}->[1],
            $quest{p2}->[0],
            $quest{p2}->[1],
        ) = m/(\d+) (\d+) (\d+) (\d+)/ if (m/^P /);
        push(@{$quest{questers}}, $1) if (m/^P\d (\w+)/);
    }
    close(QF);
}

=head FUNCTION writequestfile
    This function write the state of the current quest
    in the file.
=cut
sub writequestfile {
    return unless $opts->{writequestfile};
    open(QF,">$opts->{questfilename}") or do {
        Irpg::Irc::chanmsg("Error: Cannot open $opts->{questfilename}: $!");
        return;
    };
    # if no active quest, just empty questfile. otherwise, write it
    if (@{$quest{questers}}) {
        if ($quest{type}==1) {
            print QF "T $quest{text}\n".
                     "Y 1\n".
                     "S $quest{qtime}\n".
                     "P1 $quest{questers}->[0]\n".
                     "P2 $quest{questers}->[1]\n".
                     "P3 $quest{questers}->[2]\n".
                     "P4 $quest{questers}->[3]\n";
        }
        elsif ($quest{type}==2) {
            print QF "T $quest{text}\n".
                     "Y 2\n".
                     "S $quest{stage}\n".
                     "P $quest{p1}->[0] $quest{p1}->[1] $quest{p2}->[0] ".
                        "$quest{p2}->[1]\n".
                     "P1 $quest{questers}->[0] $rps->{$quest{questers}->[0]}{x} ".
                         "$rps->{$quest{questers}->[0]}{y}\n".
                     "P2 $quest{questers}->[1] $rps->{$quest{questers}->[1]}{x} ".
                         "$rps->{$quest{questers}->[1]}{y}\n".
                     "P3 $quest{questers}->[2] $rps->{$quest{questers}->[2]}{x} ".
                         "$rps->{$quest{questers}->[2]}{y}\n".
                     "P4 $quest{questers}->[3] $rps->{$quest{questers}->[3]}{x} ".
                         "$rps->{$quest{questers}->[3]}{y}\n";
        }
    }
    close(QF);
}

=head FUNCTION quest
    Generate a new quest.
=cut
sub quest {
    @{$quest{questers}} = grep { $rps->{$_}{online} && $rps->{$_}{level} > 39 &&
                                 time()-$rps->{$_}{lastlogin}>36000 } keys(%$rps);
    if (@{$quest{questers}} < 4) { return undef(@{$quest{questers}}); }
    while (@{$quest{questers}} > 4) {
        splice(@{$quest{questers}},int(rand(@{$quest{questers}})),1);
    }
    if (!open(Q,$opts->{eventsfile})) {
        return Irpg::Irc::chanmsg("ERROR: Failed to open $opts->{eventsfile}: $!");
    }
    my $i;
    while (my $line = <Q>) {
        chomp($line);
        if ($line =~ /^Q/ && rand(++$i) < 1) {
            if ($line =~ /^Q1 (.*)/) {
                $quest{text} = $1;
                $quest{type} = 1;
                $quest{qtime} = time() + 43200 + int(rand(43201)); # 12-24 hours
            }
            elsif ($line =~ /^Q2 (\d+) (\d+) (\d+) (\d+) (.*)/) {
                $quest{p1} = [$1,$2];
                $quest{p2} = [$3,$4];
                $quest{text} = $5;
                $quest{type} = 2;
                $quest{stage} = 1;
            }
        }
    }
    close(Q);
    if ($quest{type} == 1) {
        my $txt = "have been chosen by the gods to ".
                "$quest{text}. Quest to end in ".duration($quest{qtime}-time()).
                ".";
        Irpg::Irc::chanmsg(join(", ",(@{$quest{questers}})[0..2]).", and ".
                "$quest{questers}->[3] $txt");    
        for (my $i = 0 ; $i <= 3 ; $i++) {
            Irpg::Irc::privmsg("You, $quest{questers}->[($i+1)%4], ".
                "$quest{questers}->[($i+2)%4] and ".
                "$quest{questers}->[($i+3)%4] $txt", $quest{questers}->[$i]);
        }
    }
    elsif ($quest{type} == 2) {
        my $txt = "have been chosen by the gods to ".
                "$quest{text}. Participants must first reach [$quest{p1}->[0],".
                "$quest{p1}->[1]], then [$quest{p2}->[0],$quest{p2}->[1]].".
                ($opts->{mapurl}?" See $opts->{mapurl} to monitor their journey's ".
                "progress.":"");
        Irpg::Irc::chanmsg(join(", ",(@{$quest{questers}})[0..2]).", and ".
                "$quest{questers}->[3] $txt");
        for ($i = 0 ; $i <= 3 ; $i++) {
            Irpg::Irc::privmsg("You, $quest{questers}->[($i+1)%4], ".
                "$quest{questers}->[($i+2)%4] and ".
                "$quest{questers}->[($i+3)%4] $txt", $quest{questers}->[$i]);
        }
    }
    writequestfile();
}


=head1 FUNCTION questpencheck
    
=cut
sub questpencheck {
    my $k = shift;
    my ($quester,$player);
    for $quester (@{$quest{questers}}) {
        if ($quester eq $k) {
            my $txt = "$k\'s prudence and self-regard has brought the ".
                     "wrath of the gods upon the realm. All your great ".
                     "wickedness makes you as it were heavy with lead, ".
                     "and to tend downwards with great weight and ".
                     "pressure towards hell. Therefore have you drawn ".
                     "yourselves 15 steps closer to that gaping maw.";
            Irpg::Irc::chanmsg(Irpg::Utils::clog($txt));
            for $player (grep { $rps->{$_}{online} } keys %$rps) {
                Irpg::Irc::privmsg($txt, $player);
                my $gain = int(15 * ($opts->{rppenstep}**$rps->{$player}{level}));
                $gain = $rps->{$player}{class}->real_gain($gain);
                $rps->{$player}{pen_quest} += $gain;
                $rps->{$player}{next} += $gain;
            }
            undef(@{$quest{questers}});
            $quest{qtime} = time() + 43200; # 12 hours
        }
    }
}

sub movequesters {
    return () unless ($quest{type} == 2 && @{$quest{questers}});
    my $allgo = 1; # have all users reached <p1|p2>?
    for (@{$quest{questers}}) {
        if ($quest{stage}==1) {
            if ($rps->{$_}{x} != $quest{p1}->[0] ||
                $rps->{$_}{y} != $quest{p1}->[1]) {
                $allgo=0;
                last();
            }
        }
        else {
            if ($rps->{$_}{x} != $quest{p2}->[0] ||
                $rps->{$_}{y} != $quest{p2}->[1]) {
                $allgo=0;
                last();
            }
        }
    }
    if ($quest{stage}==1 && $allgo) {
        # all participants have reached point 1, now point 2
        Irpg::Irc::chanmsg(Irpg::Utils::clog(
            join(", ",(@{$quest{questers}})[0..2]).", ".
                 "and $quest{questers}->[3] have reached ".
                 "the first step of their long quest. May the ".
                 "Gods keep on smiling upon them."));
        for (@{$quest{questers}}) {
            Irpg::Irc::privmsg("You and your journey's companions ".
                "have reached the first step of your long quest. May the ".
                "Gods keep on smiling upon you.", $_);
        }
        $quest{stage}=2;
        $allgo=0; # have not all reached p2 yet
    }
    elsif ($quest{stage} == 2 && $allgo) {
        # all participants have reached point 2, time for reward \o/
        Irpg::Irc::chanmsg(Irpg::Utils::clog(
            join(", ",(@{$quest{questers}})[0..2]).", ".
            "and $quest{questers}->[3] have completed their ".
            "journey! 25% of their burden is eliminated. ".
            "The people can rejoice, the realm is sure a safer ".
            "place now, thanks to these heroes."));
        for (@{$quest{questers}}) {
            Irpg::Irc::privmsg("You and your companions ".
                "have completed your ".
                "journey! 25% of your burden is eliminated. ".
                "The people can rejoice, the realm is sure a safer ".
                "place now, thanks to the heroes you are.", $_);
            $rps->{$_}{next} = int($rps->{$_}{class}->real_gain(
                                    $rps->{$_}{next} * .75));
        }
        # reset the %quest hash
        undef(@{$quest{questers}});
        $quest{qtime} = time() + 21600; # next quest starts in 6 hours
        $quest{type} = 1; # probably not needed
        writequestfile();
    }
    else {
        my $phase = 'p'.(int($quest{stage} == 2) + 1);
        for (@{$quest{questers}}) {
            if (rand(100) < 1) {
                if ($rps->{$_}{x} != $quest{$phase}->[0]) {
                    $rps->{$_}{x} +=
                        (($quest{$phase}->[0] - $rps->{$_}{x}) % $opts->{mapx} <
                         ($rps->{$_}{x} - $quest{$phase}->[0]) % $opts->{mapx} ?
                            1 : -1);
                    $rps->{$_}{x} %= $opts->{mapx};
                }
                if ($rps->{$_}{y} != $quest{$phase}->[1]) {
                    $rps->{$_}{y} +=
                        (($quest{$phase}->[1] - $rps->{$_}{y}) % $opts->{mapy} <
                         ($rps->{$_}{y} - $quest{$phase}->[1]) % $opts->{mapy} ?
                            1 : -1);
                    $rps->{$_}{y} %= $opts->{mapy};
                }
            }
        }
    }
    return @{$quest{questers}};
}


sub rpcheck {
    my $rpreport = shift;
    if ($rpreport%120==0 && $opts->{writequestfile}) { writequestfile(); }
    if (time() > $quest{qtime}) {
        if (!@{$quest{questers}}) { quest(); }
        elsif ($quest{type} == 1) {
            # Time-based quest (type 1) reward
            Irpg::Irc::chanmsg(Irpg::Utils::clog(
                join(", ",(@{$quest{questers}})[0..2]).", and ".
                         "$quest{questers}->[3] have blessed the realm by ".
                         "completing their quest! 25% of their burden is ".
                         "eliminated."));
            for (@{$quest{questers}}) {
                Irpg::Irc::privmsg("You and your companions ".
                    "have blessed the realm by ".
                    "completing your quest! 25% of your burden is ".
                    "eliminated.", $_);
                $rps->{$_}{next} = int($rps->{$_}{class}->real_gain(
                                        $rps->{$_}{next} * .75));
            }
            undef(@{$quest{questers}});
            $quest{qtime} = time() + 21600;
        }
        # Grid-based quest (type 2) rewards are handled in movequesters()
        # called by Main::moveplayers()
    }
}


sub quest_info {
    my ($userhost, $usernick, $username, $source, @arg) = @_;
    if (!@{$quest{questers}}) {
        Irpg::Irc::privmsg("There is no active quest.",$source);
    }
    elsif ($quest{type} == 1) {
        Irpg::Irc::privmsg(join(", ",(@{$quest{questers}})[0..2]).", and ".
                "$quest{questers}->[3] are on a quest to ".
                "$quest{text}. Quest to complete in ".
                duration($quest{qtime}-time()).".",$source);
    }
    elsif ($quest{type} == 2) {
        Irpg::Irc::privmsg(join(", ",(@{$quest{questers}})[0..2]).", and ".
                "$quest{questers}->[3] are on a quest to ".
                "$quest{text}. Participants must first reach ".
                "[$quest{p1}->[0],$quest{p1}->[1]], then ".
                "[$quest{p2}->[0],$quest{p2}->[1]].".
                ($opts->{mapurl}?" See $opts->{mapurl} to monitor ".
                "their journey's progress.":""),$source);
    }

}


our $commands = {
    quest => {ref => \&quest_info, adm => 0, prv => 1, pub => 1,
              hlp => 'QUEST : Show status of the current quest.'}
};

1;
